# -*- coding: utf-8 -*-

"""Components to use as a sources of initial application descriptions."""

from __future__ import annotations

__all__ = [
    'RegistryKey', 'CheckConfiguredRoots', 'CheckKnownRoots', 'CheckKnownOsXRoots', 'CheckPATH',
    'CheckExecutableInWindowsRegistry', 'CheckRootInWindowsRegistry',
]

import abc
import os
import typing
from anaconda_navigator.utils.logs import logger
from .. import exceptions
from .. import import_utils
from .. import validation_utils
from . import core
from . import folders
from . import utilities

if typing.TYPE_CHECKING:
    import typing_extensions

    WindowsRegistryRoot = typing_extensions.Literal[
        'HKEY_CLASSES_ROOT',
        'HKEY_CURRENT_USER',
        'HKEY_LOCAL_MACHINE',
        'HKEY_USERS',
        'HKEY_PERFORMANCE_DATA',
        'HKEY_CURRENT_CONFIG',
    ]


# NOTE: might be replaced with `WindowsRegistryRoot.__args__` with python 3.8+
WINDOWS_REGISTRY_ROOT_KEYS: 'typing_extensions.Final[typing.Tuple[WindowsRegistryRoot, ...]]' = (
    'HKEY_CLASSES_ROOT',
    'HKEY_CURRENT_USER',
    'HKEY_LOCAL_MACHINE',
    'HKEY_USERS',
    'HKEY_PERFORMANCE_DATA',
    'HKEY_CURRENT_CONFIG',
)


class RegistryKey(typing.NamedTuple):
    """
    Information about Windows registry key, from which to get value.

    :param root: Root key to start from (e.g. :code:`'HKEY_CURRENT_USER'`, :code:`'HKEY_LOCAL_MACHINE'`.
    :param key: Path to the exact key (e.g. :code:`'app\\shell\\open\\command'`.
    :param sub_key: Subkey to get value from (might be :code:`None` to retrieve value from the root of the `key`,
                    or a name of any actual key, e.g. :code:`'ExecutablePath'`.
    :param converter: Additional modification to the key value.

                      This should be a function that accepts two arguments: value (str) and it's type identifier (int),
                      and it should return a generator of extracted values.
    """

    root: 'WindowsRegistryRoot'
    key: str
    sub_key: typing.Optional[str] = None
    converter: typing.Optional[typing.Callable[[str, int], typing.Iterator[str]]] = None

    # Possible extensions:
    #   Extend format of the key to also accept sequence of values:
    #   - strings for exact key values (just as it is right now)
    #   - predicate functions for key names (to iterate through keys, when only their format is known, but not the
    #     exact value).

    @classmethod
    def _parse_configuration(cls, configuration: typing.Mapping[str, typing.Any]) -> RegistryKey:
        """Parse configuration for :class:`~RegistryKey`."""
        validation_utils.of_type(typing.Mapping)(configuration)
        mirror: typing.Dict[str, typing.Any] = dict(configuration)

        root: 'WindowsRegistryRoot' = validation_utils.pop_mapping_item(mirror, 'root')
        validation_utils.is_str(root)
        validation_utils.of_options(*WINDOWS_REGISTRY_ROOT_KEYS)(root)

        key: str = validation_utils.pop_mapping_item(mirror, 'key')
        validation_utils.is_str(key)

        sub_key: typing.Optional[str] = mirror.pop('sub_key', None)
        validation_utils.is_str_or_none(sub_key)

        converter: typing.Optional[str] = mirror.pop('converter', None)
        validation_utils.is_str_or_none(sub_key)

        converter_value: typing.Optional[typing.Callable[..., typing.Any]]
        converter_value = import_utils.import_callable(converter, allow_none=True)

        validation_utils.mapping_is_empty()(mirror)

        return cls(root=root, key=key, sub_key=sub_key, converter=converter_value)


class CheckConfiguredRoots(core.Source, alias='check_configured_roots'):  # pylint: disable=too-few-public-methods
    """
    Get application root from the configuration.

    Multiple configuration keys might be provided as `args`.

    Configuration, from which value should be retrieved, must be provided as a keyword argument.

    If configuration section is not provided - "main" will be used by default.
    """

    __slots__ = ('__content', '__section')

    def __init__(
            self,
            *args: typing.Union[None, str, typing.Iterable[typing.Optional[str]]],
            section: str = 'applications',
    ) -> None:
        """Initialize new :class:`~CheckKnownRoots` instance."""
        self.__content: 'typing_extensions.Final[typing.Tuple[str, ...]]' = tuple(utilities.collect_str(args))
        self.__section: 'typing_extensions.Final[str]' = section

    @classmethod
    def _parse_configuration(cls, *args: typing.Any, **kwargs: typing.Any) -> core.Detector:
        """Parse configuration for this particular :class:`~Detector`."""
        with exceptions.ValidationError.with_field('args'):
            validation_utils.has_items(at_least=1)(args)
            validation_utils.each_item(validation_utils.is_str)(args)

        section: str = kwargs.pop('section', 'applications')
        validation_utils.is_str(section, field_name='section')

        validation_utils.mapping_is_empty()(kwargs)

        with validation_utils.catch_exception():
            return cls(*args, section=section)

    def __call__(self, *, context: core.DetectorContext) -> typing.Iterator[core.DetectedApplication]:
        """Iterate through detected applications."""
        content: typing.Tuple[str, ...] = self.__content
        if not content:
            content = (f'{context.app_name}_path',)

        setting: str
        for setting in content:
            root: str = context.user_configuration.get(self.__section, setting, '')
            if not root:
                continue

            root = os.path.abspath(root)
            if os.path.exists(root):
                yield core.DetectedApplication(root=root)


class CheckKnownRoots(core.Source, alias='check_known_roots'):  # pylint: disable=too-few-public-methods
    """
    Iterate through known roots, selecting only existing ones.

    Some root options might be :code:`None` - they will be skipped.
    """

    __slots__ = ('__content',)

    def __init__(self, *args: typing.Union[None, str, typing.Iterable[typing.Optional[str]]]) -> None:
        """Initialize new :class:`~CheckKnownRoots` instance."""
        self.__content: 'typing_extensions.Final[typing.Tuple[str, ...]]' = tuple(utilities.collect_str(args))

    @classmethod
    def _parse_configuration(cls, *args: typing.Any, **kwargs: typing.Any) -> core.Detector:
        """Parse configuration for this particular :class:`~Detector`."""
        new_args: typing.List[str] = []
        with exceptions.ValidationError.with_field('args'):
            validation_utils.has_items(at_least=1)(args)
            new_args.extend(map(utilities.parse_and_join, validation_utils.iterable_items(args)))  # type: ignore

        validation_utils.mapping_is_empty()(kwargs)

        with validation_utils.catch_exception():
            return cls(*new_args)

    def __call__(self, *, context: core.DetectorContext) -> typing.Iterator[core.DetectedApplication]:
        """Iterate through detected applications."""
        root: str
        for root in self.__content:
            root = os.path.abspath(root)
            if os.path.exists(root):
                yield core.DetectedApplication(root=root)


class CheckKnownOsXRoots(CheckKnownRoots, alias='check_known_osx_roots'):  # pylint: disable=too-few-public-methods
    """
    Shortcut for checking application roots of OS X applications.

    Unlike :class:`~CheckKnownRoots` - you need to provide application names instead of directory paths.
    """

    __slots__ = ()

    def __init__(self, *args: typing.Union[None, str, typing.Iterable[typing.Optional[str]]]) -> None:
        """Initialize new :class:`~CheckKnownMacRoots` instance."""
        super().__init__(*(
            utilities.join(root, arg)
            for root in (folders.FOLDERS['osx/user_applications'], folders.FOLDERS['osx/applications'])
            for arg in utilities.collect_str(args)
        ))

    @classmethod
    def _parse_configuration(cls, *args: typing.Any, **kwargs: typing.Any) -> core.Detector:
        """Parse configuration for this particular :class:`~Detector`."""
        with exceptions.ValidationError.with_field('args'):
            validation_utils.has_items(at_least=1)(args)
            validation_utils.each_item(validation_utils.is_str)(args)

        validation_utils.mapping_is_empty()(kwargs)

        with validation_utils.catch_exception():
            return cls(*args)


class CheckPATH(core.Source, alias='check_path'):  # pylint: disable=too-few-public-methods
    """
    Find path to the application executable in PATH.

    Multiple application names might be provided (like: :code:`'pycharm64.exe', 'pycharm32.exe', 'pycharm.exe'`). They
    will be checked in order they are provided, and the first match will be returned.
    """

    __slots__ = ('__content',)

    def __init__(
            self, *args: typing.Union[None, str, typing.Iterable[typing.Optional[str]]]
    ) -> None:
        """Initialize new :class:`~CheckPATH` instance."""
        self.__content: 'typing_extensions.Final[typing.Tuple[str, ...]]' = tuple(utilities.collect_str(args))

    @classmethod
    def _parse_configuration(cls, *args: typing.Any, **kwargs: typing.Any) -> core.Detector:
        """Parse configuration for this particular :class:`~Detector`."""
        with exceptions.ValidationError.with_field('args'):
            validation_utils.has_items(at_least=1)(args)
            validation_utils.each_item(validation_utils.is_str)(args)

        validation_utils.mapping_is_empty()(kwargs)

        with validation_utils.catch_exception():
            return cls(*args)

    def __call__(self, *, context: core.DetectorContext) -> typing.Iterator[core.DetectedApplication]:
        """Iterate through detected applications."""
        path: str
        for path in filter(bool, os.environ.get('PATH', '').split(os.pathsep)):
            executable: str
            for executable in self.__content:
                executable = os.path.join(path, executable)
                if not os.path.isfile(executable):
                    continue

                depth: int = 128
                while os.path.islink(executable) and (depth > 0):
                    executable = os.path.abspath(os.path.realpath(executable))
                    depth -= 1
                if depth <= 0:
                    logger.warning('recursive file link detected in system: %s', executable)
                    continue

                yield core.DetectedApplication(executable=executable)


class CheckWindowsRegistry(core.Source, metaclass=abc.ABCMeta):  # pylint: disable=too-few-public-methods
    """
    Common base for Windows registry checkers.

    It accepts list of keys that should be checked for values in registry. Each argument should be a
    :class:`~RegistryKey` instance.
    """

    __slots__ = ('__content',)

    def __init__(self, *args: RegistryKey) -> None:
        """Initialize new :class:`~CheckWindowRegistry` instance."""
        self.__content: 'typing_extensions.Final[typing.Tuple[RegistryKey, ...]]' = args

    @classmethod
    def _parse_configuration(cls, *args: typing.Any, **kwargs: typing.Any) -> core.Detector:
        """Parse configuration for this particular :class:`~Detector`."""
        new_args: typing.List[RegistryKey] = []
        with exceptions.ValidationError.with_field('args'):
            arg: typing.Any
            for arg in validation_utils.iterable_items(args):
                new_args.append(RegistryKey._parse_configuration(arg))  # pylint: disable=protected-access

        validation_utils.mapping_is_empty()(kwargs)

        with validation_utils.catch_exception():
            return cls(*new_args)

    def _values(self) -> typing.Iterator[str]:
        """
        Iterate through values that are available in the registry.

        All missing values will be ignored.
        """
        import winreg  # pylint: disable=import-error,import-outside-toplevel

        item: RegistryKey
        for item in self.__content:
            value: str
            value_type: int
            try:
                key: winreg.HKEYType  # type: ignore
                with winreg.OpenKeyEx(getattr(winreg, item.root), item.key) as key:  # type: ignore
                    value, value_type = winreg.QueryValueEx(key, item.sub_key)  # type: ignore
            except WindowsError:  # type: ignore  # pylint: disable=undefined-variable
                continue
            if item.converter is not None:
                yield from item.converter(value, value_type)
            else:
                yield value


class CheckRootInWindowsRegistry(  # pylint: disable=too-few-public-methods
    CheckWindowsRegistry,
    alias='check_root_in_windows_registry',
):
    """Check for application roots in Windows Registry."""

    __slots__ = ()

    def __call__(self, *, context: core.DetectorContext) -> typing.Iterator[core.DetectedApplication]:
        """Iterate through detected applications."""
        value: str
        for value in self._values():
            value = os.path.abspath(value)
            if os.path.isdir(value):
                yield core.DetectedApplication(root=value)


class CheckExecutableInWindowsRegistry(  # pylint: disable=too-few-public-methods
    CheckWindowsRegistry,
    alias='check_executable_in_windows_registry',
):
    """Check for paths to executable in Windows Registry."""

    def __call__(self, *, context: core.DetectorContext) -> typing.Iterator[core.DetectedApplication]:
        """Iterate through detected applications."""
        value: str
        for value in self._values():
            value = os.path.abspath(value)
            if os.path.isfile(value):
                yield core.DetectedApplication(executable=value)



# -*- coding: utf-8 -*-

"""Additional types used by NucleusAPI."""

from __future__ import annotations

__all__ = ()

import typing

if typing.TYPE_CHECKING:
    import typing_extensions

    class NucleusEnvironmentRecord(typing_extensions.TypedDict):
        """Description of the environment reported by Nucleus API."""

        id: str  # pylint: disable=invalid-name
        name: str
        yaml_ref: str
        revision: int
        created_at: str
        updated_at: str

    class NucleusEnvironmentCollection(typing_extensions.TypedDict):
        """Collection of environments reported by Nucleus API."""

        items: typing.Sequence[NucleusEnvironmentRecord]
        total: int



# -*- coding: utf-8 -*-

"""Utilities to parse repodata cache."""

from __future__ import annotations

__all__ = ['RepoDataCacheRecord', 'RepoDataCache', 'REPO_CACHE']

import contextlib
import functools
import hashlib
import itertools
import os
import re
import typing

import ujson

from anaconda_navigator.utils import logs
from anaconda_navigator.utils import workers
from anaconda_navigator.utils import worker_utils

if typing.TYPE_CHECKING:
    import typing_extensions

    RepoData = typing_extensions.TypedDict(
        'RepoData',
        {
            'packages': typing.Any,
            'packages.conda': typing.Any,
            'url': str,
        },
        total=False,
    )


class ValidCacheFiles(typing.Container[str]):  # pylint: disable=too-few-public-methods
    """
    Helper container used to validate repodata cache file names.

    It only emulates the collection so it might be used instead of lists or sets of expected names.
    """

    __slots__ = ()

    PATTERN: typing_extensions.Final[typing.Pattern[str]] = re.compile(r'^[a-z0-9]{8}\.json$')

    def __contains__(self, item: typing.Any) -> bool:
        """Check if file is a repodata cache file by its name."""
        return isinstance(item, str) and bool(self.PATTERN.fullmatch(item))


class RepoDataCacheRecord(typing.NamedTuple):
    """Single stored value in :class:`~RepoDataCache`."""

    content: 'RepoData'
    modification_time: float

    def merge(self, other: 'RepoDataCacheRecord') -> 'RepoDataCacheRecord':
        """Combine values of two cache records into a single one."""
        return RepoDataCacheRecord(
            content=typing.cast('RepoData', {**self.content, **other.content}),
            modification_time=max(self.modification_time, other.modification_time),
        )


class RepoDataCache:  # pylint: disable=too-few-public-methods
    """Helper class which caches details from repodata files."""

    __slots__ = ('__content',)

    def __init__(self) -> None:
        """Initialize new :class:`~RepoDataCache` instance."""
        self.__content: 'typing_extensions.Final[typing.Dict[str, RepoDataCacheRecord]]' = {}

    @staticmethod
    def name(url: str) -> str:
        """Get repodata cache file name from the URL of the corresponding channel."""
        if not url.endswith('/'):
            url += '/'
        return hashlib.md5(url.encode('utf-8')).hexdigest()[:8] + '.json'  # nosec

    @staticmethod
    def _normalize(content: typing.Mapping[str, typing.Any]) -> 'RepoData':
        """
        Clean and normalize parsed repodata cache file content.

        This method adapts content of the repodata cache files generated by different versions of conda to a single
        unified form.
        """
        result: 'RepoData' = {}

        # normalize the `_url` (old style) and `url` (new style) values, so they might be safely overwritten without
        # leaving any "ghost" data in collected output
        with contextlib.suppress(KeyError):
            result['url'] = content['_url']
        with contextlib.suppress(KeyError):
            result['url'] = content['url']

        with contextlib.suppress(KeyError):
            result['packages'] = content['packages']
        with contextlib.suppress(KeyError):
            result['packages.conda'] = content['packages.conda']

        return result

    @workers.Task
    def file(self, path: str, *, force: bool = False) -> RepoDataCacheRecord:
        """Read content of the single repodata cache file."""
        stream: typing.TextIO
        modification_time: float = os.path.getmtime(path)
        result: typing.Optional[RepoDataCacheRecord] = self.__content.get(path, None)
        if (result is None) or (result.modification_time < modification_time) or force:
            with open(path, 'rt', encoding='utf-8') as stream:
                result = self.__content[path] = RepoDataCacheRecord(
                    content=self._normalize(ujson.load(stream)),
                    modification_time=modification_time,
                )
        return result

    @workers.Task
    def files(self, *path: str, force: bool = False) -> RepoDataCacheRecord:
        """
        Combine content of the multiple repodata cache files.

        Each next file in the request will overwrite the collected content. I.e. the further in the argument list is,
        the higher priority of its content is. This is useful when you have multiple files of the metadata, and only one
        file with the actual content, especially since the latter one may contain metadata as well with older conda
        releases.
        """
        return worker_utils.parallel_reduce(
            RepoDataCacheRecord.merge,
            (
                self.file.worker(path=item, force=force)  # pylint: disable=no-member
                for item in path
            ),
            RepoDataCacheRecord(content={}, modification_time=0.0),
        )

    def _find_repodata(
            self,
            directories: typing.Union[str, typing.Iterable[str]],
            channels: typing.Union[None, str, typing.Iterable[str]] = None,
    ) -> typing.List[typing.List[str]]:
        """
        Detect files that contain repodata cache content.

        Method returns list of groups of files. Files in each group are sorted in order expected by
        :meth:`~RepoDataCache.files`.
        """
        result: typing.List[typing.List[str]] = []

        # prepare a set of expected repodata cache file names
        channel_filter: typing.Container[str]
        if isinstance(channels, str):
            channels = [channels]
        if channels:
            channel_filter = set(map(self.name, channels))
        else:
            channel_filter = ValidCacheFiles()

        # look for core repodata cache files in primary directories
        if isinstance(directories, str):
            directories = [directories]
        directory: str
        for directory in directories:
            directory = os.path.abspath(directory)
            if not os.path.isdir(directory):
                continue
            result.extend(
                [os.path.join(directory, name)]
                for name in os.listdir(directory)
                if name in channel_filter
            )

        # extend with extra files where necessary
        # core file should remain in the end to have the highest priority and overwrite all previous content
        #
        # code below expects each core file to have a name in form of `<hex:8>.json`
        item: typing.List[str]
        for item in result:
            # *.state.json files for conda 23.1.0+
            option: str = item[-1][:-5] + '.state.json'
            if os.path.isfile(option):
                item.insert(-1, option)

        # finish
        return result

    @workers.Task
    def collect(
            self,
            directories: typing.Union[str, typing.Iterable[str]],
            channels: typing.Union[None, str, typing.Iterable[str]] = None,
            *,
            force: bool = False,
    ) -> typing.Mapping[str, 'RepoData']:
        """Collect repodata cache from :code:`directories`, that is applicable to the :code:`channels`."""
        result: typing.Dict[str, 'RepoData'] = {}
        modification_time: typing.Dict[str, float] = {}

        thread: workers.TaskThread
        threads: typing.List[workers.TaskThread] = [
            self.files.worker(*args, force=force).thread()  # pylint: disable=no-member
            for args in self._find_repodata(directories=directories, channels=channels)
        ]
        for thread in threads:
            thread.wait()

            try:
                record: RepoDataCacheRecord = typing.cast(workers.TaskResult, thread.result).result
            except workers.TaskCanceledError:
                continue
            except Exception:  # pylint: disable=broad-except
                logs.conda_logger.exception('unable to load repodata cache file: %r', thread.call.args)
                continue

            url: typing.Optional[str] = record.content.get('url', None)
            if url is None:
                logs.conda_logger.warning('unable to parse repodata cache file: %r', thread.call.args)
                continue
            if record.modification_time > modification_time.get(url, 0.0):
                result[url], modification_time[url] = record

        return result

    @workers.Task
    def modification_time(
            self,
            directories: typing.Union[str, typing.Iterable[str]],
            channels: typing.Union[None, str, typing.Iterable[str]] = None,
    ) -> float:
        """
        Detect when repodata cache was updates last time.

        Interface is similar to the :meth:`~RepoDataCache.collect`, but returns modification time instead of the
        repodata content.
        """
        return functools.reduce(
            max,
            map(
                os.path.getmtime,
                itertools.chain.from_iterable(
                    self._find_repodata(directories=directories, channels=channels),
                )
            ),
            0.0,
        )


REPO_CACHE: 'typing_extensions.Final[RepoDataCache]' = RepoDataCache()



# -*- coding: utf-8 -*-

# pylint: disable=no-name-in-module

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""About Anaconda Navigator dialog."""

from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import (
    QAbstractItemView, QHBoxLayout, QProgressBar, QStackedWidget, QTableWidget, QTableWidgetItem, QTextEdit,
    QVBoxLayout,
)
from anaconda_navigator.api.anaconda_api import AnacondaAPI
from anaconda_navigator.api.utils import is_internet_available
from anaconda_navigator.widgets import ButtonNormal, ButtonPrimary, LabelBase, SpacerHorizontal, SpacerVertical
from anaconda_navigator.widgets.dialogs import DialogBase


class PackagesDialog(DialogBase):  # pylint: disable=too-many-instance-attributes
    """Package dependencies dialog."""

    sig_setup_ready = Signal()

    def __init__(  # pylint: disable=too-many-arguments,too-many-statements
        self,
        parent=None,
        packages=None,
        pip_packages=None,
        remove_only=False,
        update_only=False,
    ):
        """About dialog."""
        super().__init__(parent=parent)

        # Variables
        self.api = AnacondaAPI()
        self.actions = None
        self.packages = packages or []
        self.pip_packages = pip_packages or []

        # Widgets
        self.stack = QStackedWidget()
        self.table = QTableWidget()
        self.text = QTextEdit()
        self.label_description = LabelBase()
        self.label_status = LabelBase()
        self.progress_bar = QProgressBar()
        self.button_ok = ButtonPrimary('Apply')
        self.button_cancel = ButtonNormal('Cancel')

        # Widget setup
        self.text.setReadOnly(True)
        self.stack.addWidget(self.table)
        self.stack.addWidget(self.text)
        if remove_only:
            text = 'The following packages will be removed:<br>'
        else:
            text = 'The following packages will be modified:<br>'
        self.label_description.setText(text)
        self.label_description.setWordWrap(True)
        self.label_description.setWordWrap(True)
        self.label_status.setWordWrap(True)
        self.table.horizontalScrollBar().setVisible(False)
        self.table.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.table.setAlternatingRowColors(True)
        self.table.setSelectionMode(QAbstractItemView.NoSelection)
        self.table.setSortingEnabled(True)
        self._hheader = self.table.horizontalHeader()
        self._vheader = self.table.verticalHeader()
        self._hheader.setStretchLastSection(True)
        self._hheader.setDefaultAlignment(Qt.AlignLeft)
        self._hheader.setSectionResizeMode(self._hheader.Fixed)
        self._vheader.setSectionResizeMode(self._vheader.Fixed)
        self.button_ok.setMinimumWidth(70)
        self.button_ok.setDefault(True)
        self.base_minimum_width = 420
        if remove_only:
            self.setWindowTitle('Remove Packages')
        elif update_only:
            self.setWindowTitle('Update Packages')
        else:
            self.setWindowTitle('Install Packages')

        self.setMinimumWidth(self.base_minimum_width)

        # Layouts
        layout_progress = QHBoxLayout()
        layout_progress.addWidget(self.label_status)
        layout_progress.addWidget(SpacerHorizontal())
        layout_progress.addWidget(self.progress_bar)

        layout_buttons = QHBoxLayout()
        layout_buttons.addStretch()
        layout_buttons.addWidget(self.button_cancel)
        layout_buttons.addWidget(SpacerHorizontal())
        layout_buttons.addWidget(self.button_ok)

        layout = QVBoxLayout()
        layout.addWidget(self.label_description)
        layout.addWidget(SpacerVertical())
        layout.addWidget(self.stack)
        layout.addWidget(SpacerVertical())
        layout.addLayout(layout_progress)
        layout.addWidget(SpacerVertical())
        layout.addWidget(SpacerVertical())
        layout.addLayout(layout_buttons)
        self.setLayout(layout)

        # Signals
        self.button_ok.clicked.connect(self.accept)
        self.button_cancel.clicked.connect(self.reject)
        self.button_ok.setDisabled(True)

        # Setup
        self.table.setDisabled(True)
        self.update_status('Solving package specifications', value=0, max_value=0)

    def setup(self, worker, output, error):  # pylint: disable=too-many-branches,too-many-locals,too-many-statements
        """Setup the widget to include the list of dependencies."""
        if not isinstance(output, dict):
            output = {}

        packages = sorted(pkg.split('==')[0] for pkg in self.packages)
        success = output.get('success')
        error = output.get('error', '')
        exception_name = output.get('exception_name', '')
        actions = output.get('actions', [])
        prefix = worker.prefix

        if exception_name:
            message = exception_name
        else:
            # All requested packages already installed
            message = output.get('message', ' ')

        navi_deps_error = self.api.check_navigator_dependencies(actions, prefix)
        description = self.label_description.text()

        if error:
            description = 'No packages will be modified.'
            self.stack.setCurrentIndex(1)
            self.button_ok.setDisabled(True)
            if not is_internet_available():
                error = (
                    'Some of the functionality of Anaconda Navigator will be '  # pylint: disable=implicit-str-concat
                    'limited in <b>offline mode</b>.<br><br>Installation and upgrade actions will be subject to the '
                    'packages currently available on your package cache.'
                )
            self.text.setText(error)
        elif navi_deps_error:
            description = 'No packages will be modified.'
            error = 'Downgrading/removing these packages will modify Anaconda Navigator dependencies.'
            self.text.setText(error)
            self.stack.setCurrentIndex(1)
            message = 'NavigatorDependenciesError'
            self.button_ok.setDisabled(True)
        elif success and actions:
            self.stack.setCurrentIndex(0)
            # Conda 4.3.x
            if isinstance(actions, list):
                actions_link = actions[0].get('LINK', [])
                actions_unlink = actions[0].get('UNLINK', [])
            # Conda 4.4.x
            else:
                actions_link = actions.get('LINK', [])
                actions_unlink = actions.get('UNLINK', [])

            actions_link_names = {
                package['name']
                for package in actions_link
            }
            actions_unlink_names = {
                package['name']
                for package in actions_unlink
            }

            deps = set()
            deps = deps.union(actions_link_names)
            deps = deps.union(actions_unlink_names)
            deps = deps - set(packages)
            deps = sorted(list(deps))

            modified = actions_unlink_names.intersection(actions_link_names)
            modified_count = len(modified)
            plural_total_modified = self.get_plural_suffix(modified_count)

            removed = actions_unlink_names - actions_link_names
            removed_count = len(removed)
            plural_total_removed = self.get_plural_suffix(removed_count)

            installed = actions_link_names - actions_unlink_names
            installed_count = len(installed)
            plural_total_installed = self.get_plural_suffix(installed_count)

            count_total_packages = len(packages) + len(deps)
            plural_selected = self.get_plural_suffix(len(packages))

            self.table.setRowCount(count_total_packages)
            self.table.setColumnCount(5)
            self.table.sortByColumn(4, Qt.AscendingOrder)

            description = ''
            if modified:
                description = f'{description} {modified_count} package{plural_total_modified} will be modified'
            if removed:
                description = f'{description} {removed_count} package{plural_total_removed} will be removed'
            if installed:
                description = f'{description} {installed_count} package{plural_total_installed} will be installed'

            if actions_link and actions_unlink or actions_link and not actions_unlink:
                self.table.showColumn(2)
                self.table.showColumn(3)
                self.table.showColumn(4)

            elif actions_unlink and not actions_link:
                self.table.hideColumn(2)
                self.table.hideColumn(3)
                self.table.hideColumn(4)
                self.table.setHorizontalHeaderLabels(['Name', 'Unlink', 'Link', 'Channel', 'Action'])

            for row, pkg in enumerate(packages + deps):
                link_item = [
                    package
                    for package in actions_link
                    if package['name'] == pkg
                ]
                if not link_item:
                    link_item = {
                        'version': '-'.center(len('link')),
                        'channel': '-'.center(len('channel')),
                    }
                else:
                    link_item = link_item[0]

                unlink_item = [
                    package
                    for package in actions_unlink
                    if package['name'] == pkg
                ]
                if not unlink_item:
                    unlink_item = {
                        'version': '-'.center(len('link')),
                    }
                else:
                    unlink_item = unlink_item[0]

                unlink_version = str(unlink_item['version'])
                link_version = str(link_item['version'])

                item_unlink_v = QTableWidgetItem(unlink_version)
                item_link_v = QTableWidgetItem(link_version)
                item_link_c = QTableWidgetItem(link_item['channel'])
                if pkg in packages:
                    item_name = QTableWidgetItem(pkg)
                else:
                    item_name = QTableWidgetItem('*' + pkg)

                if pkg in modified:
                    action = 'Modify'
                elif pkg in removed:
                    action = 'Remove'
                elif pkg in installed:
                    action = 'Install'
                else:
                    action = 'Undetermined'
                action_item = QTableWidgetItem(action)

                items = [item_name, item_unlink_v, item_link_v, item_link_c, action_item]
                for column, item in enumerate(items):
                    item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)
                    self.table.setItem(row, column, item)

            if deps:
                message = f'<b>*</b> indicates the package is a dependency of a selected package{plural_selected}<br>'

            self.button_ok.setEnabled(True)
            self.table.resizeColumnsToContents()
            unlink_width = self.table.columnWidth(1)
            if unlink_width < 60:
                self.table.setColumnWidth(1, 60)
            self.table.setHorizontalHeaderLabels(['Name  ', 'Unlink  ', 'Link  ', 'Channel  ', 'Action  '])

        self.table.setEnabled(True)
        self.update_status(message=message)
        self.label_description.setText(description)

        # Adjust size after data has populated the table
        self.table.resizeColumnsToContents()
        width = sum(self.table.columnWidth(index) for index in range(self.table.columnCount())) + 10
        delta = (self.width() - self.table.width() + self.table.verticalHeader().width() + 10)

        new_width = width + delta

        if new_width < self.base_minimum_width:
            new_width = self.base_minimum_width

        self.setMinimumWidth(new_width)
        self.setMaximumWidth(new_width)

        self.sig_setup_ready.emit()

    @staticmethod
    def get_plural_suffix(count):  # pylint: disable=missing-function-docstring
        return 's' if count != 1 else ''

    def update_status(self, message='', value=None, max_value=None):
        """Update status of packages dialog."""
        self.label_status.setText(message)

        if max_value is None and value is None:
            self.progress_bar.setVisible(False)
        else:
            self.progress_bar.setVisible(True)
            self.progress_bar.setMaximum(max_value)
            self.progress_bar.setValue(value)



# -*- coding: utf-8 -*-

"""Collection of utility components to use by APIs."""

from __future__ import annotations

__all__ = ['normalize_certificate']

import configparser
import os
import typing

from anaconda_navigator.config import CONF


def normalize_certificate(value: typing.Union[None, bool, str]) -> typing.Union[None, bool, str]:
    """Check if certificate value is valid and fix it if required."""
    if isinstance(value, str) and (not os.path.exists(value)):
        return True
    return value


def is_internet_available() -> bool:
    """Check internet availability."""
    try:
        config_value = CONF.get('main', 'offline_mode')
    except (AttributeError, configparser.NoOptionError):
        return True
    return not bool(config_value)


def split_canonical_name(cname: str) -> typing.Tuple[str, ...]:
    """Split a canonical package name into name, version, build."""

    return tuple(cname.rsplit('-', 2))



# -*- coding: utf-8 -*-

"""
Manager for downloading resources from the web.

This manager might be used to download regular files, however - it is designed particularly for the resources, that
should be downloaded once per application launch and used continuously.

Manager is optimized to detect repeated downloads and reuse previous download results.
"""

from __future__ import annotations

__all__ = ['Keep', 'Medium', 'Status', 'Download', 'Group', 'Manager', 'manager']

import contextlib
import enum
import os
import shutil
import tempfile
import typing
import uuid

from qtpy import QtCore
import requests

from anaconda_navigator.utils import logs
from anaconda_navigator.utils import workers

if typing.TYPE_CHECKING:
    import typing_extensions

    T_co = typing.TypeVar('T_co', covariant=True)

    class ProcessDownload(typing_extensions.Protocol[T_co]):  # pylint: disable=too-few-public-methods
        """Common interface for functions to post-process download results."""

        def __call__(self, download: 'Download') -> T_co:
            """Prepare a `result` of some `download`."""

    class ProcessGroup(typing_extensions.Protocol[T_co]):  # pylint: disable=too-few-public-methods
        """Common interface for functions to post-process group download results."""

        def __call__(self, group: 'Group') -> T_co:
            """Collect a `result` of some `group`."""

    class GroupStats(typing_extensions.TypedDict):
        """Tracking stats of a group download."""

        canceled: typing.Set['Download']
        done: typing.Set['Download']
        failed: typing.Set['Download']
        requested: typing.Set['Download']
        started: typing.Set['Download']
        succeeded: typing.Set['Download']


def inject_cancel() -> bool:
    """Top-level function that injects "cancel if required" action into all downloads."""
    from anaconda_navigator import config  # pylint: disable=import-outside-toplevel

    return config.CONF.get('main', 'offline_mode', False)


def inject_kwargs() -> typing.Mapping[str, typing.Any]:
    """Top-level function that injects additional :mod:`requests` arguments into all requests."""
    from anaconda_navigator.api import client_api  # pylint: disable=import-outside-toplevel
    from anaconda_navigator.api import conda_api  # pylint: disable=import-outside-toplevel
    from anaconda_navigator.api import utils  # pylint: disable=import-outside-toplevel

    return {
        'proxies': conda_api.CondaAPI().load_proxy_config(),
        'verify': utils.normalize_certificate(client_api.ClientAPI().get_ssl(set_conda_ssl=False)),
        'timeout': 5,  # seconds
    }


def coalesce_source(sources: typing.Iterable[str]) -> str:
    """Find first file from `sources` that should be available."""
    source: str
    for source in sources:
        if os.path.isfile(source):
            return source
    raise ValueError('no usable source found')


def populate(source: str, destinations: typing.Iterable[str], overwrite: bool = False) -> None:
    """
    Copy content from `source` file into multiple `destinations`.

    By default - if any destination already exists, it would stay untouched. To change this behavior, set `overwrite` to
    :code:`True`.
    """
    destination: str
    for destination in destinations:
        destination = os.path.abspath(destination)
        if destination == source:
            continue
        if overwrite or (not os.path.isfile(destination)):
            os.makedirs(os.path.dirname(destination), exist_ok=True)
            shutil.copy2(source, destination)


def write(content: bytes, destinations: typing.Iterable[str], overwrite: bool = False) -> None:
    """
    Write binary `content` into multiple `destinations`.

    By default - if any destination already exists, it would stay untouched. To change this behavior, set `overwrite` to
    :code:`True`.
    """
    destination: str
    for destination in destinations:
        destination = os.path.abspath(destination)
        if overwrite or (not os.path.isfile(destination)):
            stream: typing.BinaryIO
            os.makedirs(os.path.dirname(destination), exist_ok=True)
            with open(destination, 'wb') as stream:
                stream.write(content)


class Keep(enum.IntEnum):
    """
    Options of keeping files being downloaded, if they already exist.

    :code:`NEVER` - would overwrite file in any case, no matter it exists or not. This is the default behavior.

    :code:`FRESH` - file would stay untouched it already exists and of the expected size.

    :code:`ALWAYS` - do not overwrite file in any case.
    """

    NEVER = enum.auto()
    FRESH = enum.auto()
    ALWAYS = enum.auto()


class Medium(str, enum.Enum):
    """
    Intermediate source to store download results in.

    :code:`MEMORY` - download all content directly into the RAM. If no download destination specified - this option
    would be used automatically, regardless of the selected medium.

    :code:`RANDOM` - use random temporary file to store intermediate download results. After download is finished - this
    file is copied into all requested destinations. This is the default behavior.

    :code:`TARGET` - write content directly into the target file. This might be an insecure option if there are clients
    that expect complete file to be available.
    """

    MEMORY = ':memory:'
    RANDOM = ':random:'
    TARGET = ':target:'


class Require(enum.IntEnum):
    """
    Control how many downloads should succeed in a :class:`Group` to treat it as a success.

    :code:`ALL` - all downloads should succeed.

    :code:`ANY` - at least one download should succeed.

    :code:`NONE` - any outcome would be treated as a success.
    """

    ALL = enum.auto()
    ANY = enum.auto()
    NONE = enum.auto()


class Status(enum.IntEnum):
    """
    Possible statuses of a download.

    :code:`PENDING` - download is not yet started.

    :code:`STARTED` - download is already in progress.

    :code:`SUCCEEDED` - download finished successfully (with a download result).

    :code:`FAILED` - download failed. There might be no download result in such case.
    """

    PENDING = enum.auto()
    STARTED = enum.auto()
    SUCCEEDED = enum.auto()
    CANCELED = enum.auto()
    FAILED = enum.auto()


DownloadT = typing.TypeVar('DownloadT', bound='Download')


DownloadIdentity = typing.Tuple[str, str, typing.Any]


class Download(QtCore.QObject):  # pylint: disable=too-many-instance-attributes
    """A single download request."""

    sig_start = QtCore.Signal(object)
    sig_succeeded = QtCore.Signal(object)
    sig_canceled = QtCore.Signal(object)
    sig_failed = QtCore.Signal(object)
    sig_done = QtCore.Signal(object)

    def __init__(self, url: str) -> None:
        """Initialize new :class:`~Download` instance."""
        super().__init__()

        self._content: typing.Optional[bytes] = None
        self._data: typing.Any = None
        self._files: typing.Set[str] = set()
        self._keep_existing: Keep = Keep.NEVER
        self._medium: typing.Union[Medium, str] = Medium.RANDOM
        self._method: str = 'GET'
        self._process: typing.Optional[ProcessDownload] = None
        self._result: typing.Any = None
        self._status: Status = Status.PENDING
        self._status_code: int = 0
        self._url: str = url
        self._worker: workers.TaskWorker = self._run.worker()  # pylint: disable=no-member

        self._worker.signals.sig_start.connect(lambda: self._start())  # pylint: disable=unnecessary-lambda
        self._worker.signals.sig_succeeded.connect(lambda _: self._succeeded())
        self._worker.signals.sig_canceled.connect(lambda _: self._canceled())
        self._worker.signals.sig_failed.connect(lambda _: self._failed())
        self._worker.signals.sig_done.connect(lambda _: self._done())

    @property
    def _identity(self) -> DownloadIdentity:  # noqa: D401
        """
        Identity of the download source.

        This is used to detect requests to the same resources - such requests should have same
        :attr:`~Download._identity`.
        """
        return self._method, self._url, self._data

    @property
    def content(self) -> bytes:  # noqa: D401
        """
        Downloaded content.

        It is unnecessary to store the content in memory in order to use this property. Content might also be stored on
        a disk - it would be fetched from it automatically.

        In case you want to read a content from a disk directly - use :meth:`Download.read` instead.
        """
        if self._content is not None:
            return self._content

        stream: typing.BinaryIO
        with self.open('rb') as stream:
            return stream.read()

    @property
    def files(self) -> typing.Tuple[str, ...]:  # noqa: D401
        """List of files into which content was downloaded."""
        return tuple(self._files)

    @property
    def result(self) -> typing.Any:  # noqa: D401
        """
        Output of the download.

        This would be set only if download is successful, and :meth:`~Download.process` function provided. Otherwise -
        it is :code:`None`.
        """
        return self._result

    @property
    def status(self) -> Status:  # noqa: D401
        """
        Current status of a download.

        Refer to :class:`Status` for available options.
        """
        return self._status

    @property
    def status_code(self) -> int:  # noqa: D401
        """HTTP status code of the resource response."""
        return self._status_code

    @property
    def usable_file(self) -> str:  # noqa: D401
        """Any file from :attr:`~Download.files` that should be available."""
        return coalesce_source(self._files)

    # ------------------------------------------------------------------------------------------------------------------

    def attach(self: DownloadT, method: str, data: typing.Any = None) -> DownloadT:
        """
        Attach data to the resource request.

        It contains of a `method` and a `data` payload. E.g. :code:`.attach('POST', b'{"message": "test"}')`.

        This method returns the self instance, which is useful for chaining calls.
        """
        if self._status != Status.PENDING:
            raise ValueError('can not change properties after download is started')

        self._method = method
        self._data = data
        return self

    def into(self: DownloadT, *files: str) -> DownloadT:
        """
        Add `files` into which resource should be downloaded.

        This method returns the self instance, which is useful for chaining calls.
        """
        if self._status == Status.FAILED:
            raise ValueError('downloaded generated no output')

        if self._status == Status.SUCCEEDED:
            if self._content is not None:
                write(self._content, self._files.difference(files), overwrite=True)
            else:
                populate(coalesce_source(self._files), self._files.difference(files), overwrite=True)

        self._files.update(map(os.path.abspath, files))
        return self

    def keep(self: DownloadT, keep_existing: Keep) -> DownloadT:
        """
        Set how existing files should be treated.

        For available options, refer to :class:`~Keep`.

        This method returns the self instance, which is useful for chaining calls.
        """
        if self._status != Status.PENDING:
            raise ValueError('can not change properties after download is started')

        self._keep_existing = keep_existing
        return self

    def process(self: DownloadT, process: 'typing.Optional[ProcessDownload]') -> DownloadT:
        """
        Set a post-process action.

        It would populate :attr:`~Download.result` if request succeeds.

        This method returns the self instance, which is useful for chaining calls.
        """
        if self._status != Status.PENDING:
            raise ValueError('can not change properties after download is started')

        self._process = process
        return self

    def via(self: DownloadT, medium: typing.Union[Medium, str]) -> DownloadT:
        """
        Set a medium to use to store intermediate download results.

        It might be a path to any file or a particular :class:`Medium` value.

        This method returns the self instance, which is useful for chaining calls.
        """
        if self._status != Status.PENDING:
            raise ValueError('can not change properties after download is started')

        self._medium = medium
        return self

    # ------------------------------------------------------------------------------------------------------------------

    @typing.overload
    def open(
            self,
            mode: 'typing_extensions.Literal["r", "rt"]' = 'rt',
            encoding: typing.Optional[str] = None,
    ) -> typing.TextIO:
        """Open resource for reading."""

    @typing.overload
    def open(
            self,
            mode: 'typing_extensions.Literal["rb"]',
    ) -> typing.BinaryIO:
        """Open resource for reading."""

    def open(self, mode='r', encoding=None):
        """Open resource for reading."""
        if 'b' not in mode:
            encoding = encoding or 'utf-8'

        return open(self.usable_file, mode, encoding=encoding)

    # ------------------------------------------------------------------------------------------------------------------

    def _canceled(self) -> None:
        """Process canceled download."""
        logs.logger.debug('download %016d: canceled', id(self))
        self._status = Status.CANCELED
        self.sig_canceled.emit(self)

    def _done(self) -> None:
        """Process completed download."""
        logs.logger.debug('download %016d: done', id(self))
        self.sig_done.emit(self)

    def _failed(self) -> None:
        """Process failed download."""
        logs.logger.debug('download %016d: failed', id(self))
        self._status = Status.FAILED
        self.sig_failed.emit(self)

    def _start(self) -> None:
        """Process started download."""
        logs.logger.debug('download %016d: started', id(self))
        self._status = Status.STARTED
        self.sig_start.emit(self)

    def _succeeded(self) -> None:
        """Process succeeded download."""
        logs.logger.debug('download %016d: succeeded', id(self))
        self._status = Status.SUCCEEDED
        self.sig_succeeded.emit(self)

    # ------------------------------------------------------------------------------------------------------------------

    @workers.Task
    def _run(self) -> None:  # pylint: disable=too-many-branches,too-many-statements
        """Perform download."""
        logs.logger.debug('download %016d: downloading %s', id(self), self._url)

        if inject_cancel():
            logs.logger.debug('download %016d: canceled due to offline mode', id(self))
            raise workers.TaskCanceledError('unable to fetch resources with offline mode enabled')

        source: typing.Optional[str] = None
        with contextlib.suppress(ValueError):
            source = coalesce_source(self._files)

        response: requests.Response
        if (source is not None) and (self._keep_existing is not Keep.NEVER):
            keep: bool = self._keep_existing is Keep.ALWAYS

            if (not keep) and (self._method == 'GET'):
                logs.logger.debug('download %016d: fetching HEAD', id(self))
                try:
                    response = requests.request('HEAD', self._url, **inject_kwargs())
                    logs.http_logger.http(response=response)
                    response.raise_for_status()
                except BaseException:  # pylint: disable=broad-except
                    logs.logger.debug('download %016d: failed to fetch HEAD', id(self), exc_info=True)
                else:
                    try:
                        keep = int(response.headers.get('Content-Length', -1)) == os.path.getsize(source)
                    except (OSError, TypeError, ValueError):
                        pass

            if keep:
                logs.logger.debug('download %016d: keeping existing content', id(self))
                populate(source, self._files, overwrite=False)
                self._finalize()
                return

        logs.logger.debug('download %016d: requesting a content', id(self))
        streaming: bool = bool(self._files) and (self._medium is not Medium.MEMORY)
        response = requests.request(self._method, self._url, data=self._data, stream=streaming, **inject_kwargs())
        logs.http_logger.http(response=response)
        self._status_code = response.status_code
        response.raise_for_status()

        if streaming:
            destination: str
            temporary: bool
            if self._medium is Medium.TARGET:
                destination = self.files[0]
                temporary = False
            elif self._medium is Medium.RANDOM:
                destination = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
                temporary = True
            else:
                destination = str(self._medium)
                temporary = destination not in self._files

            logs.logger.debug('download %016d: streaming content into %s', id(self), destination)

            os.makedirs(os.path.dirname(destination), exist_ok=True)
            response.raw.decode_content = True

            stream: typing.BinaryIO
            with open(destination, 'wb') as stream:
                shutil.copyfileobj(response.raw, stream)

            logs.logger.debug('download %016d: copying into additional files', id(self))

            populate(destination, self._files, overwrite=True)

            if temporary:
                with contextlib.suppress(OSError):
                    os.remove(destination)
        else:
            logs.logger.debug('download %016d: fetching content into memory', id(self))
            self._content = response.content

            logs.logger.debug('download %016d: writing content into files', id(self))
            write(self._content, self._files)

        self._finalize()

    def _finalize(self) -> None:
        """Apply post-processing actions to a download."""
        logs.logger.debug('download %016d: processing output', id(self))
        if self._process is not None:
            self._result = self._process(self)


GroupT = typing.TypeVar('GroupT', bound='Group')


class Group(QtCore.QObject):
    """
    Group of downloads.

    Allows managing multiple downloads simultaneously instead of each download separately.
    """

    sig_start = QtCore.Signal(object)
    sig_succeeded = QtCore.Signal(object)
    sig_canceled = QtCore.Signal(object)
    sig_failed = QtCore.Signal(object)
    sig_done = QtCore.Signal(object)

    def __init__(self, *downloads: Download) -> None:
        """Initialize new :class:`~Group` instance."""
        super().__init__()

        self._downloads: typing.Tuple[Download, ...] = downloads
        self._process: typing.Optional[ProcessGroup] = None
        self._require: Require = Require.ALL
        self._result: typing.Any = None
        self._stats: 'GroupStats' = {
            'done': set(),
            'canceled': set(),
            'failed': set(),
            'requested': set(downloads),
            'started': set(),
            'succeeded': set(),
        }
        self._status: Status = Status.PENDING

        download: Download
        for download in self._downloads:
            # pylint: disable=unnecessary-lambda
            download.sig_start.connect(lambda item: self._start(item))
            download.sig_succeeded.connect(lambda item: self._succeeded(item))
            download.sig_canceled.connect(lambda item: self._canceled(item))
            download.sig_failed.connect(lambda item: self._failed(item))
            download.sig_done.connect(lambda item: self._done(item))

            if download.status == Status.PENDING:
                continue
            self._start(download)
            if download.status == Status.STARTED:
                continue
            if download.status == Status.SUCCEEDED:
                self._succeeded(download)
            elif download.status == Status.CANCELED:
                self._canceled(download)
            elif download.status == Status.FAILED:
                self._failed(download)
            else:
                raise ValueError(f'unexpected download status: {download.status.name}')
            self._done(download)

    @property
    def downloads(self) -> typing.Tuple[Download, ...]:  # noqa: D401
        """Downloads in this group."""
        return self._downloads

    @property
    def result(self) -> typing.Any:  # noqa: D401
        """
        Output of the group.

        This would be set only if the whole group would be successful (see :meth:`~Group.require`), and
        :meth:`~Group.process` is provided. Otherwise - it is :code:`None`.
        """
        return self._result

    @property
    def status(self) -> Status:  # noqa: D401
        """
        Current status of a group download.

        Refer to :class:`Status` for available options.
        """
        return self._status

    # ------------------------------------------------------------------------------------------------------------------

    def require(self: GroupT, require: Require) -> GroupT:
        """
        Set a requirement to treat this group download a successful one.

        Refer to :class:`~Require` for available options.

        This method returns the self instance, which is useful for chaining calls.
        """
        self._require = require
        return self

    def process(self: GroupT, process: 'ProcessGroup') -> GroupT:
        """
        Set a post-process action.

        It would populate :attr:`~Group.result` if request succeeds.

        This method returns the self instance, which is useful for chaining calls.
        """
        self._process = process
        return self

    # ------------------------------------------------------------------------------------------------------------------

    def _canceled(self, download: Download) -> None:
        """Process canceled download."""
        if (not self._stats['canceled']) and (self._require is Require.ALL):
            self._status = Status.CANCELED
            self.sig_canceled.emit(self)
            self.sig_done.emit(self)

        self._stats['canceled'].add(download)

    def _done(self, download: Download) -> None:
        """Process finished download."""
        self._stats['done'].add(download)

        if (self._status is not Status.STARTED) or (self._stats['done'] != self._stats['requested']):
            return

        if (self._require is Require.NONE) or self._stats['succeeded']:
            self._finalize()
            self._status = Status.SUCCEEDED
            self.sig_succeeded.emit(self)
        elif self._stats['canceled']:
            self._status = Status.CANCELED
            self.sig_canceled.emit(self)
        else:
            self._status = Status.FAILED
            self.sig_failed.emit(self)
        self.sig_done.emit(self)

    def _failed(self, download: Download) -> None:
        """Process failed download."""
        if (not self._stats['failed']) and (self._require is Require.ALL):
            self._status = Status.FAILED
            self.sig_failed.emit(self)
            self.sig_done.emit(self)

        self._stats['failed'].add(download)

    def _start(self, download: Download) -> None:
        """Process started download."""
        if not self._stats['started']:
            self._status = Status.STARTED
            self.sig_start.emit(self)

        self._stats['started'].add(download)

    def _succeeded(self, download: Download) -> None:
        """Process succeeded download."""
        self._stats['succeeded'].add(download)

    # ------------------------------------------------------------------------------------------------------------------

    def _finalize(self) -> None:
        """Apply post-processing actions to a group."""
        if self._process is not None:
            self._result = self._process(self)


class ManagerItem(QtCore.QObject):
    """Internal collection item, that allows sharing download results between :class:`~Download` instances."""

    def __init__(self, download: Download) -> None:
        """Initialize new :class:`~ManagerItem` instance."""
        super().__init__()

        self._download: Download = download
        self._children: typing.List[Download] = []

        self._download.sig_done.connect(lambda _: self._done())

    @property
    def identity(self) -> DownloadIdentity:  # noqa: D401
        """
        Identity of the download source.

        Used for detecting downloads with the same sources.
        """
        return self._download._identity  # pylint: disable=protected-access

    def attach(self, download: Download) -> None:
        """Register another `download` which should use results of the current one."""
        logs.logger.debug('download %016d: attaching download to %016d', id(download), id(self._download))
        self._children.append(download)

        if self._download.status in (Status.SUCCEEDED, Status.CANCELED, Status.FAILED):
            self._done()

    def _apply_to(self, download: Download) -> None:
        """Apply results of the internal :class:`~Download` onto a `download`."""
        # pylint: disable=protected-access

        logs.logger.debug('download %016d: copying results from %016d', id(download), id(self._download))
        if download._status != Status.PENDING:  # pylint: disable=protected-access
            raise ValueError('download should not be started to fetch previous results')

        download._start()

        download._status_code = self._download._status_code

        if self._download._status is Status.SUCCEEDED:
            logs.logger.debug('download %016d: applying content', id(download))
            if (download._medium is Medium.MEMORY) or (not download._files):
                download._content = self._download.content

            try:
                self._download.into(*download._files)
            except ValueError:
                download._failed()
            else:
                download._finalize()
                download._succeeded()

        elif self._download._status is Status.CANCELED:
            download._canceled()

        elif self._download._status is Status.FAILED:
            download._failed()

        else:
            raise ValueError(f'unexpected status of the managed download: {self._download._status}')

        download._done()

    def _done(self) -> None:
        """Process the finish of the wrapped :class:`~Download`."""
        while self._children:
            self._apply_to(self._children.pop())


class Manager(QtCore.QObject):  # pylint: disable=too-few-public-methods
    """Central download manager."""

    def __init__(self) -> None:
        """Initialize new :class:`~Manager` instance."""
        super().__init__()

        self.__history: typing.Dict[DownloadIdentity, ManagerItem] = {}

    def execute(self, *items: typing.Union[Download, Group]) -> None:
        """Start downloads."""
        queue: typing.List[Download] = []

        item: typing.Union[Download, Group]
        for item in items:
            if isinstance(item, Download):
                queue.append(item)
            else:
                queue.extend(item.downloads)

        download: Download
        for download in queue:
            identity: DownloadIdentity = download._identity  # pylint: disable=protected-access
            try:
                parent: ManagerItem = self.__history[identity]
            except KeyError:
                self.__history[identity] = ManagerItem(download)
                download._worker.start()  # pylint: disable=protected-access
            else:
                parent.attach(download)


MANAGER: typing.Optional[Manager] = None


def manager() -> Manager:
    """Prepare singleton instance of the :class:`~Manager`."""
    global MANAGER  # pylint: disable=global-statement
    if MANAGER is None:
        MANAGER = Manager()
    return MANAGER
